class InvalidNote(Exception):
    pass


class InvalidOctave(Exception):
    pass


class InvalidPitch(Exception):
    # def __init__(self, message=''):
    #    super(InvalidPitch, self).__init__(message)
    pass


class InvalidSequenceElement(Exception):
    pass
